/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *									     *
 * Library:	libsubdiv2.5D						     *
 *									     *
 * Module:	subdiv2.5D_io.c						     *
 *									     *
 * Description:	text I/O functions					     *
 *									     *
 * Date:	20000313						     *
 *									     *
 * Author:	Francisco MORN BURGOS   (http://www.gti.ssr.upm.es./~fmb/)  *
 *		Grupo Tratamiento Imgenes, Univ. Politcnica Madrid, Spain  *
 *									     *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



#include <string.h>
#include <time.h>

#include "subdiv2.5D_priv.h"



/* * *     INPUT     * * */

/* The five following functions ("sdiReadMesh_*") differ only slightly.	*/
/* I'm not particularly proud of not having written a more "recyclable"	*/
/* code, but they would've been ugly anyway, so...			*/

static void sdiReadMesh_fmb(sdMeshPtr m, FILE *fin)
{
  int		i, n, nvtx, ivtx0, ivtx1, ivtx2;
  char		blablabla[88];
  sdVector	*vtx, *vaux, normal;
  float		r;
  sdPointPtr	*pvtx, *pvaux;
  sdBoolean	wasAlreadyThere;
  sdTrianglePtr	*mtaux, taux;

  if (!m || !fin)
    return;

  /* read number of vertices and triangles */
  for (i=0; i<3; i++)
    fgets(blablabla, sizeof(blablabla), fin);
  if (strstr(blablabla, "Sierpinski"))
    sscanf(blablabla, "# Unit %*s sphere with %d vertices and %d triangles\n",
	   &nvtx, &m->NmotherTriangles);
  else
    sscanf(blablabla, "# Unit sphere with %d vertices and %d triangles\n",
	   &nvtx, &m->NmotherTriangles);
  if (nvtx < 3)
    sdiError("sdiReadMesh_fmb: incorrect number of vertices", -5);
  if (m->NmotherTriangles < 1)
    sdiError("sdiReadMesh_fmb: incorrect number of triangles", -6);
  if (!(vtx = NEW(nvtx, sdVector)) || !(pvtx = NEW(nvtx, sdPointPtr)))
    sdiError("sdiReadMesh_fmb: not enough memory for the set of vertices", -13);
  for (i=0; i<4; i++)
    fgets(blablabla, sizeof(blablabla), fin);

  /* read and scale vertex coordinates */
  if (m->verbose > 1)
    fprintf(stderr, "\n  Reading and scaling %d vertices...  ", nvtx);
  if (m->scale > 0)	/* barycentre and scale already read or calculated */
    for (n=0, vaux=vtx; n<nvtx; n++, vaux++) {
      fscanf(fin, "%*d: {%f, %f, %f}\n", &vaux->x, &vaux->y, &vaux->z);
      DEC_VEC((*vaux), m->barycentre);
    }
  else {
    m->barycentre = sdNullVector;
    for (n=0, vaux=vtx; n<nvtx; n++, vaux++) {
      fscanf(fin, "%*d: {%f, %f, %f}\n", &vaux->x, &vaux->y, &vaux->z);
      INC_VEC(m->barycentre, (*vaux));
    }
    DIV_VEC(m->barycentre, nvtx);
    for (n=0, vaux=vtx, m->scale=0.0F; n<nvtx; n++, vaux++) {
      DEC_VEC((*vaux), m->barycentre);
      if ((r = sdmVecNorm(vaux)) > m->scale)
	m->scale = r;
    }
    m->scale = 1.0F / m->scale;
  }
  for (n=0, vaux=vtx, pvaux=pvtx; n<nvtx; n++, vaux++, pvaux++) {
    MUL_VEC((*vaux), m->scale);
    *pvaux = sdlAddPoint2VertexOT(m, &m->vertexOT, vaux, &wasAlreadyThere);
    if (m->verbose > 1 && wasAlreadyThere)
      fprintf(stderr, "\n    Warning: vertex #%d = vertex #%d.  ",
	      (*pvaux)->n, n);
    if (m->subdivScheme == SPHERE)
      (*pvaux)->normal = (*pvaux)->pos;
  }
  if (m->verbose > 1)
    fprintf(stderr, "Done.\n");

  /* read the list of original triangles */
  if (m->verbose > 1)
    fprintf(stderr, "  Reading %d triangles...  ", m->NmotherTriangles);
  fgets(blablabla, sizeof(blablabla), fin);
  if (!(m->motherTriangle = NEW(m->NmotherTriangles, sdTrianglePtr)))
    sdiError("sdiReadMesh_fmb: not enough memory for the set of triangles",
	     -13);
  if (m->progressive)
    if (!(m->motherDetailTree = NEW(m->NmotherTriangles, sdDetailTree)))
      sdiError("sdiReadMesh_fmb: not enough memory for the set of detail trees",
	       -13);
  for (n=0, mtaux=m->motherTriangle; n<m->NmotherTriangles; n++, mtaux++) {
    fscanf(fin, "%*d: {%d, %d, %d}, n = {%f, %f, %f}\n",
	   &ivtx0, &ivtx1, &ivtx2, &normal.x, &normal.y, &normal.z);
    taux = *mtaux = sdlNewLeafTriangle(m, 0);
    taux->vtx[0] = pvtx[ivtx0];
    taux->vtx[1] = pvtx[ivtx1];
    taux->vtx[2] = pvtx[ivtx2];
    pvtx[ivtx0]->t = pvtx[ivtx1]->t = pvtx[ivtx2]->t = taux;
  }
  if (m->verbose > 1)
    fprintf(stderr, "Done.\n");
  DEL(pvtx);
  DEL(vtx);
} /* sdiReadMesh_fmb */


static void sdiReadMesh_dave(sdMeshPtr m, FILE *fin)
{
  int		i, nsd, npe, te, n, nvtx, ivtx0, ivtx1, ivtx2;
  char		blablabla[88];
  sdVector	*vtx, *vaux;
  float		r;
  sdPointPtr	*pvtx, *pvaux;
  sdBoolean	wasAlreadyThere;
  sdTrianglePtr	*mtaux, taux;

  if (!m || !fin)
    return;

  /* read number of nodes, elements, space dimensions (should be 3),	*/
  /* nodes per element (3) and type of elements (2 = triangle)		*/
  for (i=0; i<5; i++)
    fgets(blablabla, sizeof(blablabla), fin);
  fscanf(fin, "%d", &nvtx); fgets(blablabla, sizeof(blablabla), fin);
  if (nvtx < 3)
    sdiError("sdiReadMesh_dave: incorrect number of vertices", -5);
  if (!(vtx = NEW(nvtx, sdVector)) || !(pvtx = NEW(nvtx, sdPointPtr)))
    sdiError("sdiReadMesh_dave: not enough memory for the set of vertices",
	     -13);
  fscanf(fin, "%d", &m->NmotherTriangles);
  fgets(blablabla, sizeof(blablabla), fin);
  if (m->NmotherTriangles < 1)
    sdiError("sdiReadMesh_dave: incorrect number of triangles", -6);
  fscanf(fin, "%d", &nsd); fgets(blablabla, sizeof(blablabla), fin);
  if (nsd != 3)
    sdiError("sdiReadMesh_dave: number of space dimensions should be 3", -4);
  fscanf(fin, "%d", &npe); fgets(blablabla, sizeof(blablabla), fin);
  if (npe != 3)
    sdiError("sdiReadMesh_dave: number of nodes per element should be 3", -11);
  fscanf(fin, "%d", &te); fgets(blablabla, sizeof(blablabla), fin);
  if (te != 2)
    sdiError("sdiReadMesh_dave: type of elements should be 2", -12);
  fgets(blablabla, sizeof(blablabla), fin);

  /* read and scale vertex coordinates */
  if (m->verbose > 1)
    fprintf(stderr, "\n  Reading and scaling %d vertices...  ", nvtx);
  if (m->scale > 0)	/* barycentre and scale already read or calculated */
    for (n=0, vaux=vtx; n<nvtx; n++, vaux++) {
      fscanf(fin, "%f %f %f\n", &vaux->x, &vaux->y, &vaux->z);
      DEC_VEC((*vaux), m->barycentre);
    }
  else {
    m->barycentre = sdNullVector;
    for (n=0, vaux=vtx; n<nvtx; n++, vaux++) {
      fscanf(fin, "%f %f %f\n", &vaux->x, &vaux->y, &vaux->z);
      INC_VEC(m->barycentre, (*vaux));
    }
    DIV_VEC(m->barycentre, nvtx);
    for (n=0, vaux=vtx, m->scale=0.0F; n<nvtx; n++, vaux++) {
      DEC_VEC((*vaux), m->barycentre);
      if ((r = sdmVecNorm(vaux)) > m->scale)
	m->scale = r;
    }
    m->scale = 1.0F / m->scale;
  }
  for (n=0, vaux=vtx, pvaux=pvtx; n<nvtx; n++, vaux++, pvaux++) {
    MUL_VEC((*vaux), m->scale);
    *pvaux = sdlAddPoint2VertexOT(m, &m->vertexOT, vaux, &wasAlreadyThere);
    if (m->verbose > 1 && wasAlreadyThere)
      fprintf(stderr, "\n    Warning: vertex #%d = vertex #%d.  ",
	      (*pvaux)->n, n);
    if (m->subdivScheme == SPHERE)
      (*pvaux)->normal = (*pvaux)->pos;
  }
  if (m->verbose > 1)
    fprintf(stderr, "Done.\n");

  /* read the list of original triangles */
  if (m->verbose > 1)
    fprintf(stderr, "  Reading %d triangles...  ", m->NmotherTriangles);
  if (!(m->motherTriangle = NEW(m->NmotherTriangles, sdTrianglePtr)))
    sdiError("sdiReadMesh_dave: not enough memory for the set of triangles",
	     -13);
  if (m->progressive)
    if (!(m->motherDetailTree = NEW(m->NmotherTriangles, sdDetailTree)))
      sdiError("sdiReadMesh_dave: not enough memory for the set of detail "
	       "trees", -13);
  for (n=0, mtaux=m->motherTriangle; n<m->NmotherTriangles; n++, mtaux++) {
    fscanf(fin, "%d %d %d\n", &ivtx0, &ivtx1, &ivtx2);
    taux = *mtaux = sdlNewLeafTriangle(m, 0);
    taux->vtx[0] = pvtx[ivtx2];
    taux->vtx[1] = pvtx[ivtx1];
    taux->vtx[2] = pvtx[ivtx0];
    pvtx[ivtx0]->t = pvtx[ivtx1]->t = pvtx[ivtx2]->t = taux;
  }
  if (m->verbose > 1)
    fprintf(stderr, "Done.\n");
  DEL(pvtx);
  DEL(vtx);
} /* sdiReadMesh_dave */


static void sdiReadMesh_tecplot(sdMeshPtr m, FILE *fin)
{
  char		blablabla[88];
  int		n, nvtx, ivtx0, ivtx1, ivtx2;
  sdVector	*vtx, *vaux;
  float		r;
  sdPointPtr	*pvtx, *pvaux;
  sdBoolean	wasAlreadyThere;
  sdTrianglePtr	*mtaux, taux;

  if (!m || !fin)
    return;

  /* read number of vertices and of triangles */
  fgets(blablabla, sizeof(blablabla), fin);
  fgets(blablabla, sizeof(blablabla), fin);
  fscanf(fin, "zone t=\"zone 1\", I=%d, J=%d", &nvtx, &m->NmotherTriangles);
  if (nvtx < 3)
    sdiError("sdiReadMesh_tecplot: incorrect number of vertices", -5);
  if (m->NmotherTriangles < 1)
    sdiError("sdiReadMesh_tecplot: incorrect number of triangles", -6);
  fgets(blablabla, sizeof(blablabla), fin);
  if (!(vtx = NEW(m->subdivScheme == SPHERE? 1 : nvtx, sdVector)) ||
      !(pvtx = NEW(nvtx, sdPointPtr)))
    sdiError("sdiReadMesh_tecplot: not enough memory for the set of vertices",
	     -13);

  /* read and scale vertex coordinates */
  if (m->verbose > 1)
    fprintf(stderr, "\n  Reading and scaling %d vertices...  ", nvtx);
  if (m->subdivScheme == SPHERE) {
    m->barycentre.x = m->barycentre.y = m->barycentre.z = 3.0F;
    for (n=0, pvaux=pvtx; n<nvtx; n++, pvaux++) {
      fscanf(fin, "%f %f %f %*f\n", &vtx->x, &vtx->y, &vtx->z);
      DEC_VEC((*vtx), m->barycentre);
      *pvaux = sdlAddPoint2VertexOT(m, &m->vertexOT, sdmNormalise(vtx),
				    &wasAlreadyThere);
      if (m->verbose > 1 && wasAlreadyThere)
	fprintf(stderr, "\n    Warning: vertex #%d = vertex #%d.  ",
		(*pvaux)->n, n);
      (*pvaux)->normal = (*pvaux)->pos;
    }
  } else {
    if (m->scale > 0)	/* barycentre and scale already read or calculated */
      for (n=0, vaux=vtx; n<nvtx; n++, vaux++) {
	fscanf(fin, "%f %f %f %*f %*f %*f\n", &vaux->x, &vaux->y, &vaux->z);
        DEC_VEC((*vaux), m->barycentre);
      }
    else {
      m->barycentre = sdNullVector;
      for (n=0, vaux=vtx; n<nvtx; n++, vaux++) {
	fscanf(fin, "%f %f %f %*f %*f %*f\n", &vaux->x, &vaux->y, &vaux->z);
        INC_VEC(m->barycentre, (*vaux));
      }
      DIV_VEC(m->barycentre, nvtx);
      for (n=0, vaux=vtx, m->scale=0.0F; n<nvtx; n++, vaux++) {
        DEC_VEC((*vaux), m->barycentre);
	if ((r = sdmVecNorm(vaux)) > m->scale)
	  m->scale = r;
      }
      m->scale = 1.0F / m->scale;
    }
    for (n=0, vaux=vtx, pvaux=pvtx; n<nvtx; n++, vaux++, pvaux++) {
      MUL_VEC((*vaux), m->scale);
      *pvaux = sdlAddPoint2VertexOT(m, &m->vertexOT, vaux, &wasAlreadyThere);
      if (m->verbose > 1 && wasAlreadyThere)
	fprintf(stderr, "\n    Warning: vertex #%d = vertex #%d.  ",
		(*pvaux)->n, n);
    }
  }
  if (m->verbose > 1)
    fprintf(stderr, "Done.\n");

  /* read the list of original triangles */
  if (!(m->motherTriangle = NEW(m->NmotherTriangles, sdTrianglePtr)))
    sdiError("sdiReadMesh_tecplot: not enough memory for the set of triangles",
	     -13);
  if (m->progressive)
    if (!(m->motherDetailTree = NEW(m->NmotherTriangles, sdDetailTree)))
      sdiError("sdiReadMesh_tecplot: not enough memory for the set of "
	       "detail trees", -13);
  if (m->verbose > 1)
    fprintf(stderr, "  Reading %d triangles...  ", m->NmotherTriangles);
  for (n=0, mtaux=m->motherTriangle; n<m->NmotherTriangles; n++, mtaux++) {
    fscanf(fin, "%d %d %d\n", &ivtx0, &ivtx1, &ivtx2);
    ivtx0--; ivtx1--; ivtx2--;
    taux = *mtaux = sdlNewLeafTriangle(m, 0);
    taux->vtx[0] = pvtx[ivtx0];
    taux->vtx[1] = pvtx[ivtx1];
    taux->vtx[2] = pvtx[ivtx2];
    pvtx[ivtx0]->t = pvtx[ivtx1]->t = pvtx[ivtx2]->t = taux;
  }
  if (m->verbose > 1)
    fprintf(stderr, "Done.\n");
  DEL(pvtx);
  DEL(vtx);
} /* sdiReadMesh_tecplot */


static void sdiReadMesh_snhc(sdMeshPtr m, FILE *fin)
{
  int		nsd, n, nvtx, ivtx0, ivtx1, ivtx2;
  sdVector	*vtx, *vaux;
  float		r;
  sdPointPtr	*pvtx, *pvaux;
  sdBoolean	wasAlreadyThere;
  sdTrianglePtr	*mtaux, taux;

  if (!m || !fin)
    return;

  /* read number of space dimensions and number of vertices */
  fscanf(fin, "%d\n", &nsd);
  if (nsd != 2 && nsd != 3)
    sdiError("sdiReadMesh_snhc: incorrect number of space dimensions", -4);
  fscanf(fin, "%d\n", &nvtx);
  if (nvtx < 3)
    sdiError("sdiReadMesh_snhc: incorrect number of vertices", -5);
  if (!(vtx = NEW(nvtx, sdVector)) || !(pvtx = NEW(nvtx, sdPointPtr)))
    sdiError("sdiReadMesh_snhc: not enough memory for the set of vertices",
	     -13);

  /* read and scale vertex coordinates */
  if (m->verbose > 1)
    fprintf(stderr, "\n  Reading and scaling %d vertices...  ", nvtx);
  if (m->scale > 0)	/* barycentre and scale already read or calculated */
    for (n=0, vaux=vtx; n<nvtx; n++, vaux++) {
      fscanf(fin, "%f %f %f\n", &vaux->x, &vaux->y, &vaux->z);
      DEC_VEC((*vaux), m->barycentre);
    }
  else {
    m->barycentre = sdNullVector;
    if (nsd == 3)
      for (n=0, vaux=vtx; n<nvtx; n++, vaux++) {
	fscanf(fin, "%f %f %f\n", &vaux->x, &vaux->y, &vaux->z);
	INC_VEC(m->barycentre, (*vaux));
      }
    else
      for (n=0, vaux=vtx; n<nvtx; n++, vaux++) {
	fscanf(fin, "%f %f\n", &vaux->x, &vaux->y);
	m->barycentre.x += vaux->x;
	m->barycentre.y += vaux->y;
      }
    DIV_VEC(m->barycentre, nvtx);
    for (n=0, vaux=vtx, m->scale=0.0F; n<nvtx; n++, vaux++) {
      DEC_VEC((*vaux), m->barycentre);
      if ((r = sdmVecNorm(vaux)) > m->scale)
	m->scale = r;
    }
    m->scale = 1.0F / m->scale;
  }
  for (n=0, vaux=vtx, pvaux=pvtx; n<nvtx; n++, vaux++, pvaux++) {
    MUL_VEC((*vaux), m->scale);
    *pvaux = sdlAddPoint2VertexOT(m, &m->vertexOT, vaux, &wasAlreadyThere);
    if (m->verbose > 1 && wasAlreadyThere)
      fprintf(stderr, "\n    Warning: vertex #%d = vertex #%d.  ",
	      (*pvaux)->n, n);
    if (m->subdivScheme == SPHERE)
      (*pvaux)->normal = (*pvaux)->pos;
  }
  if (m->verbose > 1)
    fprintf(stderr, "Done.\n");

  /* read number of triangles */
  fscanf(fin, "%d\n", &m->NmotherTriangles);
  if (m->NmotherTriangles < 1)
    sdiError("sdiReadMesh_snhc: incorrect number of triangles", -6);
  if (!(m->motherTriangle = NEW(m->NmotherTriangles, sdTrianglePtr)))
    sdiError("sdiReadMesh_snhc: not enough memory for the set of triangles",
	     -13);
  if (m->progressive)
    if (!(m->motherDetailTree = NEW(m->NmotherTriangles, sdDetailTree)))
      sdiError("sdiReadMesh_snhc: not enough memory for the set of detail "
	       "trees", -13);

  /* read the list of original triangles */
  if (m->verbose > 1)
    fprintf(stderr, "  Reading %d triangles...  ", m->NmotherTriangles);
  
  for (n=0, mtaux=m->motherTriangle; n<m->NmotherTriangles; n++, mtaux++) {
    fscanf(fin, "%d %d %d\n", &ivtx0, &ivtx1, &ivtx2);
    taux = *mtaux = sdlNewLeafTriangle(m, 0);
    taux->vtx[0] = pvtx[ivtx0];
    taux->vtx[1] = pvtx[ivtx1];
    taux->vtx[2] = pvtx[ivtx2];
    pvtx[ivtx0]->t = pvtx[ivtx1]->t = pvtx[ivtx2]->t = taux;
  }
  if (m->verbose > 1) {
    fprintf(stderr, "Done.\n");
    fflush(NULL);
  }
  DEL(pvtx);
  DEL(vtx);
} /* sdiReadMesh_snhc */


#define MAX_EDGES 88

typedef struct _PG *sdiPolygonPtr;
typedef struct _PG {
  int	nvtx;
  int	ivtx[MAX_EDGES];
} sdiPolygon;


static sdiPolygonPtr sdiReadPolygon(FILE *fin)
{
  static sdiPolygon p;
  int n, *i;

  for (n=0, i=p.ivtx; n<MAX_EDGES; n++, i++) {
    if (fscanf(fin, "%d%*c", i) < 1)	/* "%*c" skips the "," if present */
      return NULL;
    if (*i == -1)
      break;
  }
  if ((p.nvtx = n) == MAX_EDGES)
    sdiError("sdiReadPolygon: a polygon has too many edges", -9);
  return &p;
} /* sdiReadPolygon */


static void sdiReadMesh_vrml(sdMeshPtr m, FILE *fin)
{
  int		i, n, nvtx, *ivtx, ivtx0, ivtx1, ivtx2, Npolygons;
  char		blablabla[88];
  sdVector	*vtx, *vaux;
  float		r;
  sdPointPtr	*pvtx, *pvaux;
  sdBoolean	wasAlreadyThere;
  sdTrianglePtr	taux;
  sdiPolygonPtr	polygon;

  if (!m || !fin)
    return;

  /* read number of vertices and facets by making a first pass over the	*/
  /* entire file and then rewinding it -- ToDo: be (more) serious...	*/
  do
    fgets(blablabla, sizeof(blablabla), fin);
  while (!strstr(blablabla, "point ["));
  for (nvtx=0; !strchr(blablabla, ']'); nvtx++)
    fgets(blablabla, sizeof(blablabla), fin);
  if (--nvtx < 3)
    sdiError("sdiReadMesh_vrml: incorrect number of vertices", -5);
  do
    fgets(blablabla, sizeof(blablabla), fin);
  while (!strstr(blablabla, "coordIndex ["));
  Npolygons = m->NmotherTriangles = 0;
  while (polygon = sdiReadPolygon(fin)) {
    Npolygons++;
    m->NmotherTriangles += polygon->nvtx - 2;	/* polygons must be convex */
  }
  if (m->NmotherTriangles < 1)
    sdiError("sdiReadMesh_vrml: incorrect number of triangles", -6);
  if (!(vtx = NEW(nvtx, sdVector)) || !(pvtx = NEW(nvtx, sdPointPtr)))
    sdiError("sdiReadMesh_vrml: not enough memory for the set of vertices",
	     -13);
  if (!(m->motherTriangle = NEW(m->NmotherTriangles, sdTrianglePtr)))
    sdiError("sdiReadMesh_vrml: not enough memory for the set of triangles",
	     -13);
  if (m->progressive)
    if (!(m->motherDetailTree = NEW(m->NmotherTriangles, sdDetailTree)))
      sdiError("sdiReadMesh_vrml: not enough memory for the set of detail "
	       "trees", -13);
  rewind(fin);

  /* read and scale vertex coordinates */
  do
    fgets(blablabla, sizeof(blablabla), fin);
  while (!strstr(blablabla, "point ["));
  if (m->verbose > 1)
    fprintf(stderr, "\n  Reading and scaling %d vertices...  ", nvtx);
  if (m->scale > 0)	/* barycentre and scale already read or calculated */
    for (n=0, vaux=vtx; n<nvtx; n++, vaux++) {
      fscanf(fin, "%f%f%f", &vaux->x, &vaux->y, &vaux->z);
      DEC_VEC((*vaux), m->barycentre);
    }
  else {
    m->barycentre = sdNullVector;
    for (n=0, vaux=vtx; n<nvtx; n++, vaux++) {
      fscanf(fin, "%f%f%f", &vaux->x, &vaux->y, &vaux->z);
      INC_VEC(m->barycentre, (*vaux));
    }
    DIV_VEC(m->barycentre, nvtx);
    for (n=0, vaux=vtx, m->scale=0.0F; n<nvtx; n++, vaux++) {
      DEC_VEC((*vaux), m->barycentre);
      if ((r = sdmVecNorm(vaux)) > m->scale)
	m->scale = r;
    }
    m->scale = 1.0F / m->scale;
  }
  for (n=0, vaux=vtx, pvaux=pvtx; n<nvtx; n++, vaux++, pvaux++) {
    MUL_VEC((*vaux), m->scale);
    *pvaux = sdlAddPoint2VertexOT(m, &m->vertexOT, vaux, &wasAlreadyThere);
    if (m->verbose > 1 && wasAlreadyThere)
      fprintf(stderr, "\n    Warning: vertex #%d = vertex #%d.  ",
	      (*pvaux)->n, n);
    if (m->subdivScheme == SPHERE)
      (*pvaux)->normal = (*pvaux)->pos;
  }
  if (m->verbose > 1)
    fprintf(stderr, "Done.\n");

  /* read the list of original triangles */
  do
    fgets(blablabla, sizeof(blablabla), fin);
  while (!strstr(blablabla, "coordIndex ["));
  if (m->verbose > 1)
    fprintf(stderr, "  Reading %d polygons (%d triangles)...  ",
	    Npolygons, m->NmotherTriangles);
  n = 0;
  while (polygon = sdiReadPolygon(fin)) {
    ivtx = polygon->ivtx;
    /* polygons must be convex to be triangulated this way */
    for (i=0, ivtx0=*ivtx++, ivtx1=*ivtx++, ivtx2=*ivtx++;
	 i<polygon->nvtx-2; i++, ivtx1=ivtx2, ivtx2=*ivtx++) {
      taux = m->motherTriangle[n++] = sdlNewLeafTriangle(m, 0);
      taux->vtx[0] = pvtx[ivtx0];
      taux->vtx[1] = pvtx[ivtx1];
      taux->vtx[2] = pvtx[ivtx2];
      pvtx[ivtx0]->t = pvtx[ivtx1]->t = pvtx[ivtx2]->t = taux;
    }
  }
  if (m->verbose > 1)
    fprintf(stderr, "Done.\n");
  DEL(pvtx);
  DEL(vtx);
} /* sdiReadMesh_vrml */


void sdReadInitMesh(sdMeshPtr m, char *filename, sdInputFormat format)
{
  char blablabla[88];
  FILE *fin;
  int i;
  sdPointPtr p0, p1;
  sdBoolean wasAlreadyThere;
  sdEdgePtr e;
  sdVector v[3];

  if (!m)
    return;

  /* make a few guesses based on the initial mesh filename */
  if (strstr(filename, "sphere"))
    m->subdivScheme = SPHERE;
  if (format == GUESS) {
    if		(strstr(filename, ".wrl"))	format = RVRML2;
    else if	(strstr(filename, ".q"))	format = SNHC;
    else if	(strstr(filename, ".tec"))	format = TECPLOT;
    else if	(strstr(filename, ".mesh"))	format = DAVE;
    else					format = OWN;
  }

  /* eventually open details and target mesh files */
  if (m->progressive) {
    strcpy(blablabla, filename);
    strcat(blablabla, m->fDetailsBinary? ".dtb" : ".dt");
    if (!(m->fDetails = fopen(blablabla,
			      m->fDetailsIn? (m->fDetailsBinary? "rb" : "r") :
					     (m->fDetailsBinary? "wb" : "w"))))
      sdiError("sdReadInitMesh: unable to open the details file", -2);
    if (m->fDetailsIn)
      sdiReadDetailsFileHeader(m);
    else {
      strcpy(blablabla, filename);
      strcat(blablabla, ".tgt");
      if (!(fin = fopen(blablabla, "r"))) {
	if (m->verbose)
	  fprintf(stderr, "No target mesh file found: target surface will be "
			  "the unit sphere.\n");
      } else {
	/* a target mesh has to be read: do it now so that its scale	*/
	/* and barycentre can be used later for the initial mesh	*/
#ifdef DIME
	if (m->useDiMePlugin) {
	  if (m->verbose)
	    fprintf(stderr, "Reading target mesh file with DiMe plugin...\n");
	  switch (format) {
	    case SNHC:		dimeInit(blablabla, Q_FILE);	break;
	    case RVRML2:	dimeInit(blablabla, VRML_FILE);	break;
	    default:		sdiError("sdReadInitMesh: unknown target mesh "
					 "input format", -2);
	  }
	  m->scale = 1.0F / dimeScale();
	  dimeBarycenter(&m->barycentre.x, &m->barycentre.y, &m->barycentre.z);
	  if (m->verbose)
	    fprintf(stderr, "Done.\n");
	} else {
#endif
	  m->targetMesh = sdNewMesh();
	  if (m->targetMesh->verbose = m->verbose)
	    fprintf(stderr, "Reading target mesh file...  ");
	  switch (format) {
	    case OWN:		sdiReadMesh_fmb(m->targetMesh, fin);	break;
	    case DAVE:		sdiReadMesh_dave(m->targetMesh, fin);	break;
	    case TECPLOT:	sdiReadMesh_tecplot(m->targetMesh, fin);break;
	    case SNHC:		sdiReadMesh_snhc(m->targetMesh, fin);	break;
	    case RVRML2:	sdiReadMesh_vrml(m->targetMesh, fin);	break;
	    default:		sdiError("sdReadInitMesh: unknown target mesh "
					 "input format", -2);
	  }
	  fclose(fin);
	  if (m->targetMesh->verbose)
	    fprintf(stderr, "Done.\n"
			    "Building adjacency tables for target mesh...  ");
	  sdlBuildAdjacencies(m->targetMesh, FALSE);
	  if (m->targetMesh->nonManifold)
	    m->partition = FALSE;
	  if (m->targetMesh->verbose)
	    fprintf(stderr, "Done.\n"
			    "Computing triangle normals/Ks for target "
			    "mesh...  ");
	  for (i=0; i<m->targetMesh->NmotherTriangles; i++)
	    sdmFillTriangleNormalAndK(m->targetMesh->motherTriangle[i], FALSE);
	  if (m->targetMesh->verbose)
	    fprintf(stderr, "Done.\n");
	  m->barycentre = m->targetMesh->barycentre;
	  m->scale = m->targetMesh->scale;
#ifdef DIME
	}
#endif
      }
    }
  }

  /* read initial mesh */
  if (!strcmp(filename, "-"))
    fin = stdin;
  else
    if (!(fin = fopen(filename, "r")))
      sdiError("sdReadInitMesh: unable to read the initial mesh file", -2);
  if (m->verbose)
    fprintf(stderr, "Reading initial mesh file...  ");
  switch (format) {
    case OWN:		sdiReadMesh_fmb(m, fin);	break;
    case DAVE:		sdiReadMesh_dave(m, fin);	break;
    case TECPLOT:	sdiReadMesh_tecplot(m, fin);	break;
    case SNHC:		sdiReadMesh_snhc(m, fin);	break;
    case RVRML2:	sdiReadMesh_vrml(m, fin);	break;
    default:		sdiError("sdReadInitMesh: unknown input format", -2);
  }
  fclose(fin);

  /* build triangle adjacency information */
  if (m->verbose)
    fprintf(stderr, "Done.\n"
		    "Building adjacency tables for initial mesh...  ");
  sdlBuildAdjacencies(m, TRUE);

  /* compute triangle normals and areas BEFORE vertex normals */
  if (m->verbose)
    fprintf(stderr, "Done.\n"
		    "Computing triangle normals/areas for initial mesh...  ");
  for (i=0; i<m->NmotherTriangles; i++)
    sdmFillTriangleNormalAndK(m->motherTriangle[i], TRUE);
  if (m->verbose)
    fprintf(stderr, "Done.\n");

  /* build initial vertex list and ONLY THEN compute vertex normals if needed */
  sdlTransferAllVerticesFromVertexOT2VertexList(m);
  if (m->subdivScheme != SPHERE) {
    if (m->verbose)
      fprintf(stderr, "Computing vertex normals for initial mesh...  ");
    sdlComputeAllVertexNormals(m);
    if (m->verbose)
      fprintf(stderr, "Done.\n");
  }

  if (m->progressive) {
    /* build detail tree information */
    for (i=0; i<m->NmotherTriangles; i++) {
      sprintf(blablabla, "%d-", i);
      m->motherTriangle[i]->dt = m->motherDetailTree[i] =
	sdlNewLeafDetailTree(blablabla);
      m->motherDetailTree[i]->pt = m->motherTriangle[i]->vtx[0];
    }
    if (m->targetMesh && m->partition) {
      /* find correspondences between initial and target mesh vertices */
      if (!(m->targetMeshVtx = NEW(m->Npoints, sdPointPtr)))
	sdiError("sdReadInitMesh: not enough memory for the initial/target "
		 "vertices relationships", -13);
      if (m->verbose)
	fprintf(stderr, "Finding initial/target vertices relationships...  ");
      for (i=0, p0=m->firstVertex; i<m->Npoints; i++, p0=p0->next) {
	m->targetMeshVtx[i] = sdlAddPoint2VertexOT(m->targetMesh,
						   &m->targetMesh->vertexOT,
						   &p0->pos, &wasAlreadyThere);
	if (!wasAlreadyThere) {
	  fprintf(stderr, "\n  Warning: an initial mesh vertex does not belong "
			  "to the target mesh.  ");
	  m->partition = FALSE;
	}
      }
      if (m->verbose)
	fprintf(stderr, "Done.\n");
      if (m->partition) {
	/* cut target mesh into pieces */
	if (m->verbose)
	  fprintf(stderr, "Partitioning target mesh...  ");
	for (i=0, e=m->edge; i<m->Nedges; i++, e++) {
	  p0 = m->targetMeshVtx[e->vtx[0]->n];
	  p1 = m->targetMeshVtx[e->vtx[1]->n];
	  sdmBuildFrameForTriangleEdge(e->adt[0], e->e4t[0], v);
	  sdlFindPathBetweenPtsInsidePlane(m, p0, p1, v+2, NULL,
					   &m->targetMeshPath4edge[i]);
	}
	if (m->verbose)
	  fprintf(stderr, "Done.\n");
      }
    }
  }
} /* sdReadInitMesh */


static int sdiChar2SubdivSchemeAndDepth(char c, sdScheme *scheme, int *depth)
{
  *scheme = c & 0x80? MIDPOINT : BUTTERFLY;
  *depth = c & 0x7F;
  return 0;
} /* sdiChar2SubdivSchemeAndDepth */


static int sdiSwapWordsAndBytes(float *f)
{
  char *p1, *p2, aux;

  p1 = (char *) f; p2 = p1 + 3;
  aux = *p1; *p1 = *p2; *p2 = aux;
  p1++; p2--;
  aux = *p1; *p1 = *p2; *p2 = aux;
  return 0;
} /* sdiSwapWordsAndBytes */


void sdiReadDetailsFileHeader(sdMeshPtr m)
{
  char c;

  if (!m || !m->fDetails || !m->fDetailsIn)
    return;

  if (m->fDetailsBinary?
      fread(&c, sizeof(char), 1, m->fDetails) < 1
      || sdiChar2SubdivSchemeAndDepth(c, &m->subdivScheme, &m->targetDepth)
      || fread(&m->barycentre, sizeof(float), 3, m->fDetails) < 3
      || fread(&m->scale, sizeof(float), 1, m->fDetails) < 1
      || fread(&m->maxDetailN, sizeof(float), 2, m->fDetails) < 2
#ifdef LITTLE_ENDIAN
      || sdiSwapWordsAndBytes(&m->barycentre.x)
      || sdiSwapWordsAndBytes(&m->barycentre.y)
      || sdiSwapWordsAndBytes(&m->barycentre.z)
      || sdiSwapWordsAndBytes(&m->scale)
      || sdiSwapWordsAndBytes(&m->maxDetailN)
      || sdiSwapWordsAndBytes(&m->maxDetailT)
#endif
      :
      fscanf(m->fDetails, "%d %d\n%f %f %f %f\n%f %f",
	     &m->subdivScheme, &m->targetDepth,
	     &m->barycentre.x, &m->barycentre.y, &m->barycentre.z,
	     &m->scale, &m->maxDetailN, &m->maxDetailT) < 8) {
    BEEP;
    fprintf(stderr, "Warning: tried to read past end of details file.\n");
    fclose(m->fDetails);
    m->fDetails = NULL;
  }
} /* sdiReadDetailsFileHeader */


static int sdiString2Word(char *string, CODEWORD *word)
{
  char *c;
  CODEWORD mask;

  for (c=string, mask=MASK_SN; *c; c++, mask>>=1) {
    if (*c == '.') c++;
    else if (*c == '1') *word |= mask;
  }
  return 0;
} /* sdiString2Word */


void sdiReadDetail(sdMeshPtr m, CODEWORD *codedelta, sdVectorPtr delta)
{
  char bits[9*NBPDN/4];

  if (!m || !m->fDetails || !m->fDetailsIn)
    return;

  if (m->fDetailsBinary?
      fread(codedelta, sizeof(CODEWORD), 1, m->fDetails) < 1 :
      fscanf(m->fDetails, "%f %f %f %s",
	     &delta->x, &delta->y, &delta->z, bits) < 4 ||
      sdiString2Word(bits, codedelta)) {
    BEEP;
    fprintf(stderr, "Warning: tried to read past end of details file.\n");
    fclose(m->fDetails);
    m->fDetails = NULL;
  }
} /* sdiReadDetail */


sdBoolean sdiReadNextBitFromDetailsFile(sdMeshPtr m)
{
  static unsigned char c, mask = 0x01;

  if (!m || !m->fDetails || !m->fDetailsIn)
    return FALSE;

  if (m->fDetailsBinary) {
    mask >>= 1;
    if (!mask) {
      mask = 0x80;
      if (fread(&c, sizeof(c), 1, m->fDetails) < 1) {
	BEEP;
	fprintf(stderr, "Warning: tried to read past end of details file.\n");
	fclose(m->fDetails);
	m->fDetails = NULL;
	return FALSE;
      }
    }
    return (c & mask);
  } else {
    do 
      if (fscanf(m->fDetails, "%c", &c) < 1) {
	BEEP;
	fprintf(stderr, "Warning: tried to read past end of details file.\n");
	fclose(m->fDetails);
	m->fDetails = NULL;
	return FALSE;
      }
    while (c != '0' && c != '1');
    return (c == '1');
  }
} /* sdiReadNextBitFromDetailsFile */


void sdiReadDetailsFile(sdMeshPtr m, sdBoolean animate)
{
  int i;

  if (!m || !m->fDetails || !m->fDetailsIn)
    return;

  if (m->depth < m->targetDepth)
    for (i=m->depth; i<m->targetDepth; i++)
      sdSubdivAllLeafTrianglesOnce(m);

  if (m->verbose)
    fprintf(stderr, "Reading details file...  ");
  /* NB: details file header already read in "sdReadInitMesh" */
  sdSPIHTcodec(m, animate);
  fclose (m->fDetails);
  m->fDetails = NULL;
  if (m->verbose)
    fprintf(stderr, "Done.\nRepositioning all vertices...  ");
  sdRecalculateAllVertexPositions(m, 0, m->depth, NBPDN, FALSE);
  if (m->verbose)
    fprintf(stderr, "Done.\n");
} /* sdiReadDetailsFile */



/* * *     OUTPUT     * * */

static int sdiSubdivSchemeAndDepth2Char(sdScheme scheme, int depth, char *c)
{
  *c = (char) depth;
  if (scheme == MIDPOINT)
    *c |= 0x80;
  return 0;
} /* sdiSubdivSchemeAndDepth2Char */


void sdiWriteDetailsFileHeader(sdMeshPtr m)
{
  char c;

  if (!m || !m->fDetails || m->fDetailsIn)
    return;

  if (m->fDetailsBinary?
      sdiSubdivSchemeAndDepth2Char(m->subdivScheme, m->depth, &c)
      || fwrite(&c, sizeof(char), 1, m->fDetails) < 1
#ifdef LITTLE_ENDIAN
      || sdiSwapWordsAndBytes(&m->barycentre.x)
      || sdiSwapWordsAndBytes(&m->barycentre.y)
      || sdiSwapWordsAndBytes(&m->barycentre.z)
      || sdiSwapWordsAndBytes(&m->scale)
      || sdiSwapWordsAndBytes(&m->maxDetailN)
      || sdiSwapWordsAndBytes(&m->maxDetailT)
#endif
      || fwrite(&m->barycentre, sizeof(float), 3, m->fDetails) < 3
      || fwrite(&m->scale, sizeof(float), 1, m->fDetails) < 1
      || fwrite(&m->maxDetailN, sizeof(float), 2, m->fDetails) < 2
      :
      fprintf(m->fDetails, "%1d %1d\n", m->subdivScheme, m->depth) < 4
      || fprintf(m->fDetails, "%+E %+E %+E %+E\n", m->barycentre.x,
		 m->barycentre.y, m->barycentre.z, m->scale) < 56
      || fprintf(m->fDetails, "%+E %+E\n", m->maxDetailN, m->maxDetailT) < 28) {
    BEEP;
    fprintf(stderr, "Warning: could not write to details file.\n");
    fclose(m->fDetails);
    m->fDetails = NULL;
  }
} /* sdiWriteDetailsFileHeader */


static int sdiWord2String(CODEWORD word, char *string)
{
  CODEWORD mask;
  char *c;
  int i;

  for (mask=MASK_SN, c=string, i=0; mask; mask>>=1, c++, i++) {
#ifdef BIT_SHUFFLING
    if (i == 3 || i == 3 + NBPDN-NBPDT)
#else
    if (mask & MASK_ST1 || mask & MASK_ST2)
#endif
      *c++ = '.';
    *c = (word & mask)? '1' : '0';
  }
  return *c = '\0';
} /* sdiWord2String */


void sdiWriteDetail(sdMeshPtr m, CODEWORD *codedelta, sdVectorPtr delta)
{
  static sdBoolean firstCall = TRUE;
  char bits[NBPD+3];

  if (!m || !m->fDetails || m->fDetailsIn)
    return;

  if (firstCall && !m->fDetailsBinary)
    firstCall = FALSE;
  if (m->fDetailsBinary?
      fwrite(codedelta, sizeof(CODEWORD), 1, m->fDetails) < 1 :
      sdiWord2String(*codedelta, bits) ||
      fprintf(m->fDetails, "%+E %+E %+E %s\n",
	      delta->x, delta->y, delta->z, bits) < 42 + sizeof(bits)) {
    BEEP;
    fprintf(stderr, "Warning: could not write to details file.\n");
    fclose(m->fDetails);
    m->fDetails = NULL;
  }
} /* sdiWriteDetail */


static void sdiWriteNextBit2DetailsFile(sdMeshPtr m, sdBoolean b,
					sdBoolean flush)
{
  static unsigned char c = 0, mask = 0x80;
  static int i = 0;

  if (b) c |= mask;
  mask >>= 1;
  if (!mask || flush && mask != 0x80) {
    if (fwrite(&c, sizeof(c), 1, m->fDetails) < 1) {
      BEEP;
      fprintf(stderr, "Warning: could not write to details file.\n");
      fclose(m->fDetails);
      m->fDetails = NULL;
    }
    c = 0; mask = 0x80; i++;
  }
  if (flush && m->verbose)
    fprintf(stderr, "(%d bytes)  ", i);
} /* sdiWriteNextBit2DetailsFile */


void sdiWriteNextBitToDetailsFile(sdMeshPtr m, sdBoolean b)
{
  static int i = 0;

  if (!m || !m->fDetails || m->fDetailsIn)
    return;

  if (m->fDetailsBinary)
    sdiWriteNextBit2DetailsFile(m, b, FALSE);
  else {
    if (!(i % 64))
      fprintf(m->fDetails, "\n");
    else if (!(i % 8))
      fprintf(m->fDetails, ".");
    if (fprintf(m->fDetails, "%c", b? '1' : '0') < 1) {
      BEEP;
      fprintf(stderr, "Warning: could not write to details file.\n");
      fclose(m->fDetails);
      m->fDetails = NULL;
    }
    i++;
  }
} /* sdiWriteNextBitToDetailsFile */


void sdiWriteDetailsFile(sdMeshPtr m)
{
  if (!m || !m->fDetails || m->fDetailsIn)
    return;

  if (m->verbose)
    fprintf(stderr, "Writing details file...  ");
  sdiWriteDetailsFileHeader(m);
  sdlCalculateMaxDetailsInAllDetailTrees(m);
  sdSPIHTcodec(m, FALSE);
  if (m->fDetailsBinary)
    sdiWriteNextBit2DetailsFile(m, FALSE, TRUE);
  else
    fprintf(m->fDetails, "\n");
  fclose(m->fDetails);
  m->fDetails = NULL;
  if (m->verbose)
    fprintf(stderr, "Done.\n");
} /* sdgWriteDetailsFile */


void sdiPrintPoint(sdMeshPtr m, sdPointPtr p,
		   sdOutputFormat format, sdBoolean normal)
{
  if (!m || !p)
    return;

  switch (format) {
    case USELESS:
      printf("Point #%5d: ", p->n);
      printf("pos=(%+5.3f, %+5.3f, %+5.3f)", p->pos.x, p->pos.y, p->pos.z);
      if (normal)
	printf("; normal=(%+5.3f, %+5.3f, %+5.3f)",
	       p->normal.x, p->normal.y, p->normal.z);
      printf(".\n");
      break;

    case MATHEMATICA:
      /* NOP: no vertices list in Mathematica format */
      break;

    case VRML2:
      printf("        %+E %+E %+E\n", p->pos.x, p->pos.y, p->pos.z);
      break;

    default:
      sdiError("sdiPrintPoint: unknown output format", -3);
  }
} /* sdiPrintPoint */


static void sdiPrintAllPointsInVertexList(sdMeshPtr m, sdOutputFormat format,
					  sdBoolean normal)
{
  sdPointPtr pt;

  if (!m)
    return;

  for (pt=m->firstVertex; pt; pt=pt->next)
    sdiPrintPoint(m, pt, format, normal);
} /* sdiPrintAllPointsInVertexList */


void sdiPrintTriangle(sdMeshPtr m, sdTrianglePtr t,
		      sdOutputFormat format, sdBoolean normal)
{
  int spc;
  sdDetailTree dt;

  if (!m)
    return;

  switch (format) {
    case USELESS:
      if (!t)
	printf("Triangle #   -1\n");
      else {
	for (spc=0; spc<t->depth; spc++) printf("  ");
	printf("Triangle #%5d (%c%c%d):\n",
	       t->n, t->inv? 'i' : 'n', t->subdiv? 'm' : 's', t->depth);
	for (spc=0; spc<t->depth; spc++) printf("  ");
	printf("  vtx={%5d, %5d, %5d};  adj={%5d, %5d, %5d}.\n",
	       t->vtx[0]? t->vtx[0]->n : -1,
	       t->vtx[1]? t->vtx[1]->n : -1,
	       t->vtx[2]? t->vtx[2]->n : -1,
	       t->adt[0]? t->adt[0]->n : -1,
	       t->adt[1]? t->adt[1]->n : -1,
	       t->adt[2]? t->adt[2]->n : -1);
	if (m->progressive && !t->inv) {
	  for (spc=0; spc<t->depth; spc++) printf("  ");
	  printf("  t->dt->id=");
	  if (!(dt = t->dt))
	    printf("<error>.\n");
	  else {
	    while (dt->son && dt->son[0]) {
	      printf("\"%s\", ", dt->id);
	      dt = dt->son[0];
	    }
	    printf("\"%s\".\n", dt->id);
	  }
	}
	if (normal) {
	  for (spc=0; spc<t->depth; spc++) printf("  ");
	  printf("  n=(%+5.3f, %+5.3f, %+5.3f); a=%1.3E.\n",
		 t->normal.x, t->normal.y, t->normal.z, t->K);
	}
      }
      break;

    case MATHEMATICA:
      if (t && !t->subdiv)
	printf("Polygon[{{%f,%f,%f},{%f,%f,%f},{%f,%f,%f}}],\n",
	       t->vtx[0]->pos.x, t->vtx[0]->pos.y, t->vtx[0]->pos.z,
	       t->vtx[1]->pos.x, t->vtx[1]->pos.y, t->vtx[1]->pos.z,
	       t->vtx[2]->pos.x, t->vtx[2]->pos.y, t->vtx[2]->pos.z);
      break;

    case VRML2:
      if (!t)
	BEEP;
      else if (!t->subdiv) {
	if (normal)
	  printf("        %+E %+E %+E\n",
		 t->normal.x, t->normal.y, t->normal.z);
	else
	  printf("        %5d %5d %5d -1\n",
		 t->vtx[0]->n, t->vtx[1]->n, t->vtx[2]->n);
      }
      break;

    default:
      sdiError("sdiPrintTriangle: unknown output format", -3);
      break;
  }
} /* sdiPrintTriangle */


static void sdiPrintAllTrianglesInMT(sdMeshPtr m, sdTrianglePtr t,
				     sdOutputFormat format, sdBoolean normal)
{
  int spc, i;

  if (!m || !t)
    return;

  switch (format) {
    case USELESS:
      sdiPrintTriangle(m, t, USELESS, normal);
      if (t->subdiv) {
	for (spc=0; spc<t->depth; spc++) printf("  ");
	printf("{\n");
	for (i=0; i<4; i++)
	  sdiPrintAllTrianglesInMT(m, t->son[i], USELESS, normal);
	for (spc=0; spc<t->depth; spc++) printf("  ");
	printf("}\n");
      }
      break;

    case MATHEMATICA:
    case VRML2:
      if (t->subdiv)
	for (i=0; i<4; i++)
	  sdiPrintAllTrianglesInMT(m, t->son[i], format, normal);
      else
	sdiPrintTriangle(m, t, format, normal);
      break;

    default:
      sdiError("sdiPrintAllTrianglesInMT: unknown output format", -3);
  }
} /* sdiPrintAllTrianglesInMT */


static void sdiPrintAllTriangles(sdMeshPtr m,
				 sdOutputFormat format, sdBoolean normal)
{
  int i;

  if (!m)
    return;

  for (i=0; i<m->NmotherTriangles; i++)
    sdiPrintAllTrianglesInMT(m, m->motherTriangle[i], format, normal);
} /* sdiPrintAllTriangles */


#ifdef WIN32
#  define WHOAMI "subdiv"
#else
#  define WHOAMI "libsubdiv2.5D"
#endif

void sdPrintMesh(sdMeshPtr m, sdOutputFormat format)
{
  time_t t;
  char blablabla[88];

  if (!m)
    return;

  time(&t);
  strftime(blablabla, sizeof(blablabla), "%Y%m%d%H%M%S", localtime(&t));

  switch (format) {
    case USELESS:
      /* print header */
      printf("Mesh {\t/* created by %s (%s) */\n\n", WHOAMI, blablabla);
      /* print vertex list */
      printf("Vertices {\n");
      sdiPrintAllPointsInVertexList(m, USELESS, FALSE);
      printf("} /* Vertices */\n\n");
      /* print triangle list and tail */
      printf("Triangles {\n");
      sdiPrintAllTriangles(m, USELESS, FALSE);
      printf("} /* Triangles */\n\n");
      printf("} /* Mesh */\n\n");
      break;

    case MATHEMATICA:
      /* no header, no vertex list: print directly triangle list and tail */
      printf("facets=Graphics3D[{\n");
      sdiPrintAllTriangles(m, MATHEMATICA, FALSE);
      printf("},AspectRatio->1,Axes->True,RenderAll->False]\n");
      break;

    case VRML2:
      /* print header */
      printf("#VRML V2.0 utf8\n");
      printf("\n");
      printf("WorldInfo {\n");
      printf("  title \"Untitled (yeah, right, yet another conceptual art "
	     "masterpiece).\"\n");
      printf("  info [ \"Created by %s (%s).\" ]\n", WHOAMI, blablabla);
      printf("}\n");
      printf("\n");
      printf("NavigationInfo { type \"EXAMINE\" }\n");
      printf("\n");
      printf("Background { skyColor 0.3 0.0 0.1 }\n");
      printf("\n");
      printf("Transform { children [\n");
      printf("  Shape {\n");
      printf("    appearance Appearance {\n");
      printf("      material Material {\n");
      printf("        diffuseColor  0.2 1.0 0.8\n");
      printf("        specularColor 0.5 0.1 0.5\n");
      printf("        shininess     0.88\n");
      printf("      } # Material\n");
      printf("    } # Appearance\n");
      printf("    geometry IndexedFaceSet {\n");
      /* print vertex list */
      printf("      coord Coordinate { point [\n");
      sdiPrintAllPointsInVertexList(m, VRML2, FALSE);
      printf("      ] } # point, Coordinate\n");
      /* print triangle list */
      printf("      coordIndex [\n");
      sdiPrintAllTriangles(m, VRML2, FALSE);
      printf("      ] # coordIndex\n");
      /* print normal list */
      printf("      normalPerVertex FALSE\n");
      printf("      normal Normal { vector [\n");
      sdiPrintAllTriangles(m, VRML2, TRUE);
      printf("      ] } # vector, Normal\n");
      /* print tail */
      printf("    } # IndexedFaceSet\n");
      printf("  } # Shape\n");
      printf("] } # children, Transform\n");
      printf("\n");
      break;

    default:
      sdiError("sdPrintMesh: unknown output format", -3);
  }

  fflush(stdout);
} /* sdPrintMesh */


static void sdiPrintDetailTree(sdDetailTree dt)
{
  int indent, i;
  char bits[NBPD+3], bitsD[NBPD+3], bitsL[NBPD+3];

  if (!dt)
    return;

  indent = 2 * strlen(strchr(dt->id, '-')) - 1;
  for (i=0; i<indent; i++) printf("  ");
  printf("DetailTree \"%s\" {\n", dt->id);
  for (i=0; i<indent; i++) printf("  ");
  printf("  pt #%5d; d = (%E, %E, %E)\n",
	 dt->pt->n, dt->pt->delta.x, dt->pt->delta.y, dt->pt->delta.z);
  for (i=0; i<indent; i++) printf("  ");
  sdiWord2String(dt->pt->codedelta, bits);
  printf("  c = %s\n", bits);
  for (i=0; i<indent; i++) printf("  ");
  sdiWord2String(dt->D, bitsD);
  printf("  D = %s\n", bitsD);
  for (i=0; i<indent; i++) printf("  ");
  sdiWord2String(dt->L, bitsL);
  printf("  L = %s\n", bitsL);
  if (dt->son) {
    for (i=0; i<indent; i++) printf("  ");
    printf("  sons {\n");
    for (i=0; i<4; i++)
      sdiPrintDetailTree(dt->son[i]);
    for (i=0; i<indent; i++) printf("  ");
    printf("  }\n");
  }
  for (i=0; i<indent; i++) printf("  ");
  printf("}\n");
} /* sdiPrintDetailTree */


void sdiPrintAllDetailTrees(sdMeshPtr m)
{
  int i;

  if (!m || !m->progressive)
    return;

  sdlCalculateMaxDetailsInAllDetailTrees(m);
  printf("DetailTrees {\n");
  for (i=0; i<m->NmotherTriangles; i++)
    sdiPrintDetailTree(m->motherTriangle[i]->dt);
  printf("} /* DetailTrees */\n\n");
} /* sdiPrintAllDetailTrees */



/* * *     _THE_ FUNCTION     * * */

int sdiError(char *msg, int rc)	/* "int" so that it can be used in macroes */
{
  /* error codes:							*/
  /*  -1: incompatibilities between progressive and adaptive modes	*/
  /*  -2: unknown input format						*/
  /*  -3: unknown output format						*/
  /*  -4: incorrect number of space dimensions				*/
  /*  -5: incorrect number of vertices					*/
  /*  -6: incorrect number of triangles					*/
  /*  -7: not a manifold mesh (an edge belongs to three triangles)	*/
  /*  -8: yuyu								*/
  /*  -9: a polygon has too many edges					*/
  /* -10: tried to delete a ghost point					*/
  /* -11: incorrect number of nodes per element				*/
  /* -12: incorrect element type					*/
  /* -13: not enough memory						*/
  fprintf(stderr, "\n\n%s: error!!!\n  %s.\a\n\n", WHOAMI, msg);
  exit(rc);
  return rc;	/* to avoid silly warnings */
} /* sdiError */

